import os
from datetime import datetime

try:
    from urllib.parse import quote_plus
except ImportError:
    from urllib import quote_plus

from scrapy.spiders import CrawlSpider, Rule
from scrapy.linkextractors import LinkExtractor

from scrapy_wayback_machine import WaybackMachineMiddleware


class MirrorSpider(CrawlSpider):
    name = 'mirror_spider'
    handle_httpstatus_list = [404]

    def __init__(self, domains, directory, allow=(), deny=(), unix=False):
        self.directory = directory
        self.unix = unix
        self.rules = (
            Rule(LinkExtractor(allow=allow, deny=deny), callback='save_page'),
        )

        # parse the allowed domains and start urls
        self.allowed_domains = []
        self.start_urls = []
        for domain in domains:
            url_parts = domain.split('://')
            unqualified_url = url_parts[-1]
            url_scheme = url_parts[0] if len(url_parts) > 1 else 'http'
            full_url = '{0}://{1}'.format(url_scheme, unqualified_url)
            bare_domain = unqualified_url.split('/')[0]
            self.allowed_domains.append(bare_domain)
            self.start_urls.append(full_url)

        super().__init__()

    def parse_start_url(self, response):
        # scrapy doesn't call the callbacks for the start urls by default,
        # this overrides that behavior so that any matching callbacks are called
        for rule in self._rules:
            if rule.link_extractor._link_allowed(response):
                if rule.callback:
                    rule.callback(response)

    def save_page(self, response):
        # ignore 404s
        if response.status == 404:
            return

        # make the parent directory
        url_parts = response.url.split('://')[1].split('/')
        if os.name == 'nt':
            url_parts = [quote_plus(url_part) for url_part in url_parts]
        parent_directory = os.path.join(self.directory, *url_parts)
        os.makedirs(parent_directory, exist_ok=True)

        # construct the output filename
        time = response.meta['wayback_machine_time']
        if self.unix:
            filename = '{0}.snapshot'.format(time.timestamp())
        else:
            filename = '{0}.snapshot'.format(time.strftime(WaybackMachineMiddleware.timestamp_format))
        full_path = os.path.join(parent_directory, filename)

        # write out the file
        with open(full_path, 'wb') as f:
            f.write(response.body)



import argparse
from pkg_resources import get_distribution

from scrapy.crawler import CrawlerProcess
from scrapy.settings import Settings

from .mirror_spider import MirrorSpider


def main():
    # configure the settings for the crawler and spider
    args = parse_args()
    config = {
        'domains': args.domains,
        'directory': args.output,
        'allow': args.allow,
        'deny': args.deny,
        'unix': args.unix,
    }
    settings = Settings({
        'USER_AGENT': (
            'Wayback Machine Scraper/{0} '
            '(+https://github.com/sangaline/scrapy-wayback-machine)'
        ).format(get_distribution('wayback-machine-scraper').version),
        'LOG_LEVEL': 'DEBUG' if args.verbose else 'INFO',
        'DOWNLOADER_MIDDLEWARES': {
            'scrapy_wayback_machine.WaybackMachineMiddleware': 5,
        },
        'AUTOTHROTTLE_ENABLED': True,
        'AUTOTHROTTLE_DEBUG': args.verbose,
        'AUTOTHROTTLE_START_DELAY': 1,
        'AUTOTHROTTLE_TARGET_CONCURRENCY': args.concurrency,
        'WAYBACK_MACHINE_TIME_RANGE': (getattr(args, 'from'), args.to),
    })

    # start the crawler
    process = CrawlerProcess(settings)
    process.crawl(MirrorSpider, **config)
    process.start()


def parse_args():
    formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(formatter_class=formatter, description=(
        'Mirror all Wayback Machine snapshots of one or more domains '
        'within a specified time range.'
    ))
    parser.add_argument('domains', metavar='DOMAIN', nargs='+', help=(
        'Specify the domain(s) to scrape. '
        'Can also be a full URL to specify starting points for the crawler.'
    ))
    parser.add_argument('-o', '--output', metavar='DIRECTORY', default='website', help=(
        'Specify the domain(s) to scrape. '
        'Can also be a full URL to specify starting points for the crawler.'
    ))
    parser.add_argument('-f', '--from', metavar='TIMESTAMP', default='10000101', help=(
        'The timestamp for the beginning of the range to scrape. '
        'Can either be YYYYmmdd, YYYYmmddHHMMSS, or a Unix timestamp.'
    ))
    parser.add_argument('-t', '--to', metavar='TIMESTAMP', default='30000101', help=(
        'The timestamp for the end of the range to scrape. '
        'Use the same timestamp as `--from` to specify a single point in time.'
    ))
    parser.add_argument('-a', '--allow', metavar='REGEX', default=(), help=(
        'A regular expression that all scraped URLs must match.'
    ))
    parser.add_argument('-d', '--deny', metavar='REGEX', default=(), help=(
        'A regular expression to exclude matched URLs.'
    ))
    parser.add_argument('-c', '--concurrency', default=10.0, help=(
        'Target concurrency for crawl requests.'
        'The crawl rate will be automatically adjusted to match this target.'
        'Use values less than 1 to be polite and higher values to scrape more quickly.'
    ))
    parser.add_argument('-u', '--unix', action='store_true', help=(
        'Save snapshots as `UNIX_TIMESTAMP.snapshot` instead of '
        'the default `YYYYmmddHHMMSS.snapshot`.'
    ))
    parser.add_argument('-v', '--verbose', action='store_true', help=(
        'Turn on debug logging.'
    ))

    return parser.parse_args()



from setuptools import find_packages, setup

description = ('A command-line utility for scraping '
               'Wayback Machine snapshots from archive.org.')
long_description = description + \
        (' For further details, '
         'please see the code repository on github: '
         'https://github.com/sangaline/wayback-machine-scraper')


setup(
    name='wayback-machine-scraper',
    version='1.0.8',
    author='Evan Sangaline',
    author_email='evan@intoli.com',
    description=description,
    license='ISC',
    keywords='archive.org scrapy scraper waybackmachine',
    url="https://github.com/sangaline/wayback-machine-scraper",
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'wayback-machine-scraper = wayback_machine_scraper.__main__:main',
        ],
    },
    long_description=long_description,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Scrapy',
        'Topic :: Utilities',
        'License :: OSI Approved :: ISC License (ISCL)',
    ],
    install_requires=[
        'cryptography',
        'scrapy',
        'scrapy-wayback-machine',
        'twisted',
    ]
)
